(define-module (alterator ajax)

	       :use-module (vhttpd)
	       :use-module (alterator d)
	       :use-module (alterator ahttpd)
	       :use-module (alterator woo)
	       :use-module (alterator algo)
	       :use-module (alterator str)
	       :use-module (alterator base64)
	       :use-module (alterator homovector)

	       :use-module (alterator plist)
	       :use-module (srfi srfi-1)

	       :re-export (;;
			   message?
			   message-header
			   message-cookie
			   message-method
			   message-uri
			   message-query
			   message-body

			   decode-url-component
			   encode-url-component)
	       :export (;; for internal use
			form-callback
			make-ajax-response
			ui-callback
			make-ui-response
			;;client part
			form-bind
			form-set-timeout
			form-replace
			form-update-enum
			form-update-value
			form-update-value-list
			form-value
			form-value-list
			form-update-visibility
			form-update-activity
			;;error handling
			catch/message
			form-warning
			form-error
			form-woo-error
			form-type-error
			;;cookie support
			form-update-cookie
			form-cookie
			;; simple UI processing
			ui-blob
			ui-blob?
			ui-file
			ui-file?
			ui-replace
			ui-replace?
			))

;;; response generation

(define *response* (make-fluid))

(define *uri* (make-fluid))
(define *msg* (make-fluid))

(define (make-response)
  '())

(define (response-append fmt . args)
  (let ((response (fluid-ref *response*)))
    (or (null? response) (pair? response) (error "function called out of ajax context"))
    (fluid-set! *response*
		(cons (apply format #f fmt args) response))))

(define (response-text)
  (string-concatenate (reverse (fluid-ref *response*))))

(define (make-ajax-response uri callbackname msg)
  (with-fluids ((woo-gate d-query)
		(*response* (make-response))
		(*uri* uri)
		(*msg* msg))
	       (form-callback callbackname uri)
	       (make-string-response 200
				     "text/javascript;charset=utf-8"
				     (response-text))))

(define (make-ui-response uri msg)
  (with-fluids ((woo-gate d-query)
		(*response* #f)
		(*uri* uri)
		(*msg* msg))
	       (let ((answer (ui-callback uri)))
		 (cond
		   ((ui-blob? answer)
		    (let* ((file-name (ui-blob-filename answer))
			   (response  (make-string-response 200
							    (ui-blob-type answer)
							    (u8vector->ISO-8859-1-string (ui-blob-data answer)))))
		      (and file-name
			   (message-header-set! response
						"content-disposition"
						(format #f "attachment;filename=~A" file-name)))
		      response))
		   ((ui-file? answer)
		    (make-file-response (ui-file-path answer) ""))
		   ((ui-replace? answer)
		    (make-redirect-response (ui-replace-path answer)))
		   (else
		     ;; ignore other answers
		     #f)))))

;;; callbacks resolving

(define *callbacks* (make-hash-table 20))
(define *random-state* (seed->random-state (current-time)))
(define *random-max* 100000000)

(define (form-callback callbackname uri)
  (let* ((callback (callback-proc uri callbackname)))
    (and (procedure? callback) (callback))))

(define (callback-proc url name)
  (or (static-callback url name)
      (dynamic-callback url name)))

(define (dynamic-callback url name)
  (let ((url (string-trim url #\/)))
    (dynamic-require `(ui ,@(map string->symbol (string-split url #\/)) ajax)
		     (string->symbol name))))

(define (callback-key url name)
  (string-append url name))

(define (callback-name proc)
  (let ((real-name (procedure-name proc)))
    (if real-name
      (symbol->string real-name)
      (number->string (random *random-max* *random-state*) 16))))

(define (static-callback url name)
   (hash-ref *callbacks* (callback-key url name) #f))

(define (callback-id proc)
  (let ((name (callback-name proc)))
    (hash-set! *callbacks*
                (callback-key (fluid-ref *uri*) name)
                proc)
    name))

;;; json

(define (real-plist? x)
  (and (plist? x)
       (or (null? x)
	   (not-pair? (car x)))))

(define (->json obj)
  (cond
    ((and (number? obj) (zero? (imag-part obj))) (number->json obj))
    ((string? obj) (string->json obj))
    ((symbol? obj) (symbol->json obj))
    ((boolean? obj) (boolean->json obj))
    ((real-plist? obj)  (plist->json obj))
    ((list? obj) (list->json obj))
    (else (error "unsupported object for serialization" obj))))

(define (number->json num)
  (number->string (exact->inexact num)))

(define (string->json str)
  (string-append "\""
		 (string-quote (lambda (ch)
				 (case ch
				   ((#\newline) "\\n")
				   ((#\\) "\\\\")
				   ((#\") "\\\"")
				   (else (string ch))))
			       str)
		 "\""))

(define (symbol->json sym)
  (string->json (symbol->string sym)))

(define (boolean->json b)
  (string->json (if b "#t" "#f")))

(define (plist->json lst)
  (string-append "{"
		 (string-join
		   (plist-map (lambda(x y)
				(format #f "~A:~A" (->json x) (->json y)))
			      lst)
		   ",")
		 "}"))

(define (list->json lst)
  (string-append "["
		 (string-join (map ->json lst) ",")
		 "]"))

;;; user functions for forms

(define (form-bind name event proc)
  (response-append "form_bind(~A,~A,~A);~%"
		   (->json name)
		   (->json event)
		   (->json (callback-id proc))))

(define (form-set-timeout proc timeout)
  (response-append "form_set_timeout(~A,~A);~%"
		   (->json (callback-id proc))
		   (->json (inexact->exact (round (* timeout 1000))))))

(define (plist->url-args lst)
  (string-join
    (plist-map
      (lambda(x y) (format #f "~A=~A"
			   (encode-url-component (->string x))
			   (encode-url-component (->string y))))
      lst)
    "&"))

(define (form-replace url . url-args)
  (response-append "form_replace(~A);~%"
		   (->json
		     (string-append url
				    (if (pair? url-args)
				      (if (string-index url #\?) "&" "?")
				      "")
				    (plist->url-args url-args)))))

(define (form-update-enum name cmdlist)
  (response-append "form_update_enum(~A,~A);~%"
		   (->json name)
		   (->json (map cdr cmdlist))))

(define (form-update-value name value)
  (response-append "form_update_value(~A,~A);~%"
		   (->json name)
		   (->json value)))

(define (form-update-value-list namelist cmd)
    (for-each (lambda(name)
		(form-update-value name (woo-get-option cmd (string->symbol name))))
	      namelist))

(define (form-value-list . namelist)
  (apply backend-args (message-woo-args (fluid-ref *msg*)) namelist))

(define (form-value name . default)
  (apply cond-assoc (string->symbol name)
	 (message-woo-args (fluid-ref *msg*))
	 default))

(define (form-update-visibility name status)
  (let ((namelist (if (pair? name) name (list name))))
    (for-each
      (lambda(name)
	(response-append "form_update_visibility(~A,~A)~%"
			 (->json name)
			 (if status "true" "false")))
      namelist)))

(define (form-update-activity name status)
  (let ((namelist (if (pair? name) name (list name))))
    (for-each
      (lambda(name)
	(response-append "form_update_activity(~A,~A)~%"
			 (->json name)
			 (if status "true" "false")))
      namelist)))

;;; simple error handling

(define (form-error fmt . args)
  (response-append "form_error(~A)~%"
		   (->json (apply format #f fmt args))))

(define (form-warning fmt . args)
  (response-append "form_warning(~A)~%"
		   (->json (apply format #f fmt args))))

(define form-woo-error form-error)

(define (form-type-error reasonlist)
  (response-append "form_type_error(~A)~%"
		   (->json (alist->plist reasonlist))))

(define (catch/message proc)
  (catch
    #t
    proc
    (lambda (key . args)
      (case key
	((woo-error)  (form-woo-error (car args)))
	((type-error) (form-type-error (car args)))
	(else (form-error "key=~S, args=~S" key args)))
      #f)))

(define (form-update-cookie name value)
  (response-append "form_update_cookie(~A,~A)~%"
		   (->json name)
		   (->json value)))

(define (form-cookie name)
  (message-cookie (fluid-ref *msg*) name))


;;; Basic UI processing

(define (ui-callback uri)
  (let ((proc (dynamic-callback uri "ui")))
    (and (procedure? proc)
	 (proc))))

(define (ui-blob name data . content-type)
  (cons* 'blob
         (base64-string->u8vector (woo-get-option data (string->symbol name)))       ;; content
         (if (pair? content-type) (car content-type) "application/octet-stream")     ;; mime-type
	 (and (pair? content-type) (pair? (cdr content-type)) (cadr content-type)))) ;; filename

(define (ui-blob? data)
  (and (pair? data)
       (eq? (car data) 'blob)))

(define (ui-blob-data data)
  (cadr data))

(define (ui-blob-type data)
  (caddr data))

(define (ui-blob-filename data)
  (cdddr data))

(define (ui-file name data)
  (cons 'file (woo-get-option data (string->symbol name))))

(define (ui-file? data)
  (and (pair? data)
       (eq? (car data) 'file)))

(define (ui-file-path data)
  (cdr data))

(define (ui-replace url . url-args)
  (cons 'replace
	(string-append url
		       (if (pair? url-args)
			 (if (string-index url #\?) "&" "?")
			 "")
		       (plist->url-args url-args))))

(define (ui-replace? data)
  (and (pair? data)
       (eq? (car data) 'replace)))

(define (ui-replace-path data)
  (cdr data))

